from collections.abc import Sequence, MutableSequence, Set, MutableSet
from typing import Iterable

class Maillon:
    def __init__(self, valeur, suivant):
        self._valeur = valeur
        self._suivant = suivant

    @property
    def valeur(self):
        return self._valeur

    @property
    def suivant(self):
        return self._suivant

    def change_valeur(self, valeur):
        self._valeur = valeur

    def change_suivant(self, suivant):
        self._suivant = suivant


    def get_maillon(self, index):
        if index == 0:
            return self
        if self.suivant:
            return self.suivant.get_maillon(index-1)
        raise IndexError()

    def get(self, index):
        return self.get_maillon(index).valeur

class ListeChainee(Sequence):
    def __init__(self, valeurs=None):
        maillon = None
        if not valeurs:
            self._premier_maillon = None
        else:
            for valeur in reversed(valeurs):
                maillon = Maillon(valeur, maillon)
            self._premier_maillon = maillon 

    @property
    def premier_maillon(self):
        if self._premier_maillon is None:
            raise IndexError(0)
        return self._premier_maillon

    def __len__(self):
        n = 0
        try:
            maillon = self.premier_maillon
        except IndexError:
            return 0
        while maillon is not None:
            n += 1
            maillon = maillon.suivant
        return n
             
    def __getitem__(self, index):
        return self.premier_maillon.get(index)

    def __repr__(self):
        return "[" + "⇝".join(map(str, self)) + "]"


class ListeChaineeMutable(MutableSequence, ListeChainee):
    def __setitem__(self, indice, value):
        if self.premier_maillon:
            self.premier_maillon.get_maillon(indice).change_valeur(valeur)

    def __delitem__(self, indice):
        if indice == 0:
            self._premier_maillon = self.premier_maillon.get_maillon(1)
            return
        maillon = self.premier_maillon.get_maillon(indice-1)
        if maillon.suivant is None:
            raise IndexError(indice)
        maillon.change_suivant(maillon.suivant.suivant)

    def insert(self, indice, valeur):
        try:
            maillon = self.premier_maillon.get_maillon(indice)
            maillon_suivant = maillon.suivant
            nouveau_maillon = Maillon(valeur, maillon.suivant)
            maillon.change_suivant(nouveau_maillon)
            # maillon.change_suivant(Maillon(valeur, maillon.suivant))
        except IndexError:
            if indice == 0 and self._premier_maillon is None:
                self._premier_maillon = Maillon(valeur, None)
            else:
                dernier_maillon = self.premier_maillon.get_maillon(indice - 1)
                dernier_maillon.change_suivant(Maillon(valeur, None))


class Ensemble(Set):
    """ Via liste, mauvaise complexité """
    def __init__(self, valeurs: Iterable):
        self._liste = list()
        for a in valeurs:
            if a not in self._liste:
                self._liste.append(a)
    def __contains__(self, valeur):
        return valeur in self._liste
    def __iter__(self):
        return iter(self._liste)
    def __len__(self):
        return len(self._liste)


class EnsembleHash(Set):
    def __init__(self, valeurs: Iterable):
        self._count = 0
        self.build(valeurs, 2*len(valeurs))

    def build(self, valeurs, n: int):
        n = len(valeurs)
        L = [[] for i in range(n)]
        for v in valeurs:
            K = L[hash(v) % (n)]
            if v not in K:
                K.append(v)
                self._count += 1
        self._data = L

    def __contains__(self, valeur):
        K = self._data[hash(valeur) % (len(self._data))]
        return valeur in K

    def __iter__(self):
        for K in self._data: 
            yield from K

    def __len__(self):
        return self._count 

    def __repr__(self):
        return f"<" + f",".join(map(str, self)) + ">"

class EnsembleMutableHash(EnsembleHash, MutableSet):
    ratio = 1.7

    def add(self, key):
        if key not in self:  # self.__contains__(key)
            K = self._data[hash(key) % (len(self._data))]
            K.append(key)
            self._count += 1
            if len(self) > (self.ratio * (len(self._data))):
                self.resize()

    def resize(self):
        self.build(self, 2* len(self._data))

    def discard(self, key):
        if key in self:  # self.__contains__(key)
            K = self._data[hash(key) % (len(self._data))]
            K.remove(key)
            self._count -= 1
